__author__ = "Andrea Rubbi"
# ──────────────────────────────────────────────────────────────────────────────
from sklearn.decomposition import PCA
import numpy as np
import pandas as pd
# ──────────────────────────────────────────────────────────────────────────────
# ─── PCA N COMPONENTS ─────────────────────────────────────────────────────────
def pca(distance_matrix, n_components, metadata=None):
    """embed distance_matrix in n_components with Principal Coordinate Analysis

    Args:
        distance_matrix (pandas.DataFrame): distance_matrix
        n_components (int): number of desired components
        metadata (pandas.DataFrame, optional): metadata of elements. Defaults to None.

    Returns:
        components (numpy.array): principal coordinates(components) of distance matrix 
    """
    pca = PCA(n_components=n_components)
    components = pca.fit_transform(distance_matrix)
    total_var = pca.explained_variance_ratio_.sum() * 100
    pd.DataFrame(components).to_csv('./PCA_Embedding.csv')

    return components


import os, sys
from rich import print
import readline
import rlcompleter
readline.parse_and_bind("tab: complete")
# getting the name of the directory
current = os.path.dirname(os.path.realpath(__file__))
 
# Getting the parent directory name
parent = os.path.dirname(current)
 
# adding the parent directory to
# the sys.path.
sys.path.append(parent)

def usage():
    print('[bold green]PEAR | Interactive Mode')
    print('[green]⣿⣿⣿⣿⣿⣿⣿⣿⠿⣟⠉⡿⠿⣿⣿⣿⣿⣿⣿⣿[bold white] -- Controls --',
          '[green]⣿⣿⣿⣿⡿⣿⢉⢳⠴⣞⠉⡷⢥⡏⡙⡿⢿⣿⣿⣿[bold white] 1 --> see status',
          '[green]⣿⣿⡋⢻⡤⣼⠉⢯⡆⣞⠙⣧⣢⠏⠪⣣⢦⡛⠹⣿[bold white] 2 --> calculate distances',
          '[green]⣿⣿⠓⢻⣄⣼⠋⢷⡠⡽⠚⣉⣤⡞⢚⢦⢢⠟⠹⣿[bold white] 3 --> embed distances',
          '[green]⣿⣿⢓⢻⡄⡼⠗⢃⣂⡒⠻⣧⣂⡿⠚⣨⢨⡓⠻⣿[bold white] 4 --> plot embeddings',
          '[green]⣿⣿⠗⢎⢄⠂⠾⣯⣂⡽⢓⢆⢔⠐⠿⣇⢅⡗⠻⣿[bold white] 5 --> add set to collection',
          '[green]⣿⣿⠖⢯⡡⣹⠗⣤⣉⠛⠶⣏⢌⡿⠲⡌⢌⢞⠼⣿[bold white] 6 --> get subset',
          '[green]⣿⣿⣮⣾⡉⣹⠦⣞⡉⡽⠶⡌⠌⡮⠲⣏⢩⣷⣼⣿[bold white]',
          '[green]⣿⣿⣿⣿⣿⣿⣤⣞⢉⣳⠥⣏⠍⣧⣵⣿⣿⣿⣿⣿[bold white] 7 --> exit',
          '[green]⣿⣿⣿⣿⣿⣿⣿⣿⣿⣷⣤⣿⣿⣿⣿⣿⣿⣿⣿⣿[bold white] 8 --> see list of controls', sep = '\n')
    
    return 0



def exit_program():
    print('- Leaving PEAR -')
    exit()

def calculate_distances(SET):
    methods = ['hashrf', 'hashrf_weighted', 'days_RF']
    while True:
        method = int(input("Method (1:hashrf - 2:weighted hashrf - 3:day\'s rf): "))
        if method in (1,2,3): break
        else: print('[bold orange]Please select 1, 2 or 3')
    
    SET.calculate_distances(methods[method - 1])
    return 0
    
def embedding(SET):
    while True:
        method = int(input("Method (1:PCoA - 2:t-SNE): "))
        if method in (1,2): break
        else: print('[bold orange]Please select either 1 or 2')
        
    while True:
        dimensions = int(input("Number of dimensions: "))
        if dimensions in (2,3): break
        else: print('[bold orange]Please select either 2 or 3')
        
    if method == 1:
        SET.embed('pca', dimensions)
    elif method == 2:
        SET.embed('tsne', dimensions)
    return 0

def plotting(SET):
    methods = ['pca', 'tsne']
    while True:
        method = int(input("Method (1:PCoA - 2:t-SNE): "))
        if method in (1,2): break
        else: print('[bold orange]Please select either 1 or 2')
        
    while True:
        dimensions = int(input("Number of dimensions: "))
        if dimensions in (2,3): break
        else: print('[bold orange]Please select either 2 or 3')
    
    while True:
        show = input("Show figure (y/n): (default is no)  ")
        if show in ('y', 'n', 'Y', 'N', 'Yes', 'No', 'yes', 'no', ''): break
        else: print('[bold orange]Please select valid value')
        
    if dimensions == 2:
        fig = SET.plot_2D(methods[method - 1], save = True)
    elif dimensions == 3:
        fig = SET.plot_3D(methods[method - 1], save = True)
    
    if show in ('y', 'Y', 'Yes', 'yes'): fig.show()
    return 0

def add_set():
    global SET
    while True:
        filename = input("Specify file name: ")
        try: 
            open(filename, 'r').close()
            break
        except: print('[bold orange]Cannot find the specified file')
    return filename

def get_subset(SET):
    n_required = int(input("Number of elements in subset: "))
    SET.get_subset(n_required)
    return 0



def interact(control):
    try:
        Actions = {
            1 : "print(SET)",
            2 : "interactive.calculate_distances(SET)",
            3 : "interactive.embedding(SET)",
            4 : "interactive.plotting(SET)",
            5 : "SET = set_collection(SET) + set_collection(tree_set(interactive.add_set()))", #"interactive.add_set()",
            6 : "get_subset(SET)",
            7 : "interactive.exit_program()",
            8: "interactive.usage()",
        }
        return Actions[control]
    except: return "print('Select a key from the list!')"
    


__author__ = 'Andrea Rubbi'

import plotly.graph_objects as go
import plotly.express as px
import pandas as pd
import numpy as np
from ipywidgets import widgets
import builtins
import matplotlib.colors as mcolors
from pylab import cm
import matplotlib
import random
import warnings


def plot_embedding(data, metadata, dimensions, save=False, name_plot='Tree_embedding', static=False, plot_meta = 'SET-ID', plot_set = None, select = False, same_scale = False):
    """Plot embedding of distance matrix - in 2D or 3D

    Args:
        data (pandas.DataFrame): embedding of distance matrix 
        metadata (pandas.DataFrame): metadata of tree_set or set_collection
        dimensions (int): number of dimensions - either 2 or 3
        save (bool, optional): save plot in pdf format. Defaults to False.
        name_plot (str, optional): name of plot. Defaults to 'Tree_embedding'.
        static (bool, optional): if True, returns a less interactive format of plot. Defaults to False.
        plot_meta (str, optional): defines the meta-feature used to color the points. Defaults to 'SET-ID'.
        plot_set (list or str, optional): sets in set_collection to be plotted. Defaults to None.
        select (bool, optional): if True, generates widgets that allow to show or hide uo to 16 set traces. Defaults to False.

    Returns:
       image: plot with related widgets - interactive or static format
    """
    
    # ─── Format Controls ──────────────────────────────────────────────────
    assert dimensions <= 3 and dimensions > 1, 'Please select either 2 or 3 dimensions for the plot'  
    assert plot_meta in metadata.columns, f'Could not find {plot_meta} in metadata'
    assert 'SET-ID' in metadata.columns, 'Could not find SET-ID in metadata - please redefine SET-ID or reset the tree_set/set_collection and subsequently define additional metadata'
    
    # ─── plot_set Selects The Tree Set To Be Displayed ────────────────────
    if type(plot_set) == type(None): plot_set = np.unique(metadata['SET-ID'])
    else: plot_set = list(plot_set)

    idxs_plot_set = [Tset in plot_set for Tset in metadata['SET-ID']]
    data, metadata = data[idxs_plot_set], metadata[idxs_plot_set]
    
    # ─── Define Color Map And Mapping Array For Each Metadata ─────────────
    metadata_colors = dict()
    for meta in metadata.columns: # for each metadata column
        elements = np.unique(metadata[meta]) # unique elements
        color_plot = np.array(metadata[meta].values) # get meta column
        
        colorscales_cmap = ['Accent', 'Blues', 'BrBG', 'BuGn', 'BuPu', 'CMRmap', 'Dark2',
                       'GnBu', 'Greens', 'Greys', 'OrRd', 'Oranges', 'PRGn', 'Paired',
                       'Pastel1', 'Pastel2', 'PiYG', 'PuBu', 'PuBuGn', 'PuOr', 'PuRd',
                       'Purples', 'RdBu', 'RdGy', 'RdPu', 'RdYlBu', 'RdYlGn', 'Reds',
                       'Set1', 'Set2', 'Set3', 'Spectral', 'Wistia', 'YlGn', 'YlGnBu',
                       'YlOrBr', 'YlOrRd', 'afmhot', 'autumn', 'binary', 'bone', 'brg',
                       'bwr', 'cividis', 'cool', 'coolwarm', 'copper', 'cubehelix', 
                       'flag', 'gist_earth', 'gist_gray', 'gist_heat', 'gist_ncar', 
                       'gist_stern', 'gist_yarg', 'gnuplot', 'gnuplot2', 'gray', 'hot',
                       'hsv', 'inferno', 'jet', 'magma', 'nipy_spectral', 'ocean', 'pink',
                       'plasma', 'prism', 'rainbow', 'seismic', 'spring', 'summer', 
                       'tab10', 'tab20', 'tab20b', 'tab20c', 'terrain', 'turbo', 'twilight',
                       'twilight_shifted', 'viridis', 'winter']
        # colorscales_cmap contains all colorscales compatible with cm.get_cmap function
        # we consider only jet as it contains enough-contrasting colors | other colormaps may be considered
        # n future updates :: TODO::
        colorscales = ['jet']                             
        
        colorscales_GO = ['aggrnyl', 'agsunset', 'algae', 'amp', 'armyrose', 'balance',
             'blackbody', 'bluered', 'blues', 'blugrn', 'bluyl', 'brbg',
             'brwnyl', 'bugn', 'bupu', 'burg', 'burgyl', 'cividis', 'curl',
             'darkmint', 'deep', 'delta', 'dense', 'earth', 'edge', 'electric',
             'emrld', 'fall', 'geyser', 'gnbu', 'gray', 'greens', 'greys',
             'haline', 'hot', 'hsv', 'ice', 'icefire', 'inferno', 'jet',
             'magenta', 'magma', 'matter', 'mint', 'mrybm', 'mygbm', 'oranges',
             'orrd', 'oryel', 'oxy', 'peach', 'phase', 'picnic', 'pinkyl',
             'piyg', 'plasma', 'plotly3', 'portland', 'prgn', 'pubu', 'pubugn',
             'puor', 'purd', 'purp', 'purples', 'purpor', 'rainbow', 'rdbu',
             'rdgy', 'rdpu', 'rdylbu', 'rdylgn', 'redor', 'reds', 'solar',
             'spectral', 'speed', 'sunset', 'sunsetdark', 'teal', 'tealgrn',
             'tealrose', 'tempo', 'temps', 'thermal', 'tropic', 'turbid',
             'turbo', 'twilight', 'viridis', 'ylgn', 'ylgnbu', 'ylorbr',
             'ylorrd']
        
        # single colors for interpretability
        colorscales_GO = ['blues', 'darkmint', 'gray', 'greens', 'greys',
             'hot', 'ice', 'magenta', 'mint', 'oranges', 'peach',
             'purples', 'reds']
        
        colorscale_cont = random.sample(colorscales, 1)[0]
        cmap = cm.get_cmap(colorscale_cont, max(20,len(elements)))
        # here we generate a color map --> if there are more than 1000 unique elements,
        # we assume that the variable is some sort of common parameter. e.g. likelihood, step, parsimony...
        cont_colorsc = False
        if len(elements) > 10: cont_colorsc = True 
        # else, we presume this is some kind of difference between sets - e.g. different tree_set
        # we generate random colors to maximize the visual division
        else: col_list = [mcolors.rgb2hex(cmap(i)[:3]) for i in range(cmap.N)]
        
        # for each single element in meta, we generate a common color_plot array
        # where each element is encoded as distinctive number
        for i, elem in enumerate(elements):
            idx = color_plot == elem
            color_plot[idx] = i
        
        metadata_colors[f'{meta}_color_plot'] = list()
        metadata_colors[f'{meta}_color_map'] = list()
        
        Sets = np.unique(metadata['SET-ID'])
        for i, SetID in enumerate(Sets):
            idx = metadata['SET-ID'] == SetID
            color_plot_set = color_plot[idx]
            if cont_colorsc: col_list = random.sample(colorscales_GO, 1)[0] if not same_scale else 'jet'
            color = col_list
            if len(np.unique(color_plot_set)) == 1: color = random.sample([mcolors.rgb2hex(cmap(i)[:3]) for i in range(cmap.N)], 10)
            metadata_colors[f'{meta}_color_plot'].append(color_plot_set.tolist())
            metadata_colors[f'{meta}_color_map'].append(color)
            
        
    # initialize go.Figure()
    fig = go.Figure()
    
    # widget for the modification of metadata coloring
    meta_widget = widgets.Dropdown(
        options=list(metadata.columns),
        value=plot_meta,
        description='Metadata:',
        )
    
    # widget to save plot current view as pdf
    save_pdf = widgets.Button(
            description='Save plot as PDF',
            button_style='danger',
            layout= widgets.Layout(width="150px"),)
    
    # define response_meta to change color of points with different metadata selections
    def response_meta(change):
        traces_start_at = 0
        if dimensions == 2: traces_start_at = 1
        with fig.batch_update():
            for i in range(nUnique):
                fig.data[i + traces_start_at].marker['color'] = metadata_colors[f'{meta_widget.value}_color_plot'][i]
                fig.data[i + traces_start_at].marker['colorscale'] = metadata_colors[f'{meta_widget.value}_color_map'][i]

    # bind widget to function response_metadata
    meta_widget.observe(response_meta, names="value")
    
    # define save_pdf_funct to save pdf
    def save_pdf_func(b):
        fig.write_image(name_plot + '.pdf')
    
    # bind widget to function save_pdf_func
    save_pdf.on_click(save_pdf_func)
    
    
    # ─── 3D Plot ──────────────────────────────────────────────────────────
    if dimensions == 3: # if input dimension is 3
        assert data.shape[1] == 3, 'Embed distance_matrix in 3D before requesting a 3D plot'
        
        # add a scatter3d trace for each tree_set - i.e. SET-ID unique value
        Sets, nSetID = np.unique(metadata['SET-ID'], return_counts=True)
        # number of sets
        nUnique = len(Sets)
        for i, SetID in enumerate(Sets):
            idx = metadata['SET-ID'] == SetID
            fig.add_trace(go.Scatter3d(x=data[idx,0],
                                y=data[idx,1],
                                z=data[idx,2], mode= 'markers', 
                                showlegend=False,
                                visible=True,
                                marker_color=metadata_colors[f'{plot_meta}_color_plot'][i],
                                text=metadata['SET-ID'].values[idx],
                                opacity = 0.7,
                                marker=dict(
                                colorscale=metadata_colors[f'{plot_meta}_color_map'][i],
                                size = 10,
                                line=dict(
                                        #color='MediumPurple',
                                        width=10,
                                        color = "rgba( 30, 30, 30, 0.3)",
                                    ),
                                ),
                                ))
        
        # defines the layout of the plot    
        fig.update_layout(
            width=900,
            height=900 + 10*nUnique,
            autosize=True,
            margin=dict(t=0, b=100, l=0, r=0),
            template="seaborn",
            plot_bgcolor='rgba(0,0,0,0)',
            paper_bgcolor='rgba(0,0,0,0)',
            #hovermode="x unified",
        )

        # updates 3D scene options
        fig.update_scenes(
            aspectratio=dict(x=1, y=1, z=1),
            aspectmode="manual"
        )
        
        
        buttons_ID = list()
        # buttons_ID are switch on-off buttons that allow to show or hyde specific tree_sets | activated by select
        # TODO : transform double buttons in single toggle button --> specify the hyde-button behaviours in arg2
        if select:
            if nUnique > 16:
                warnings.warn('Selection not available for number of sets > 16 - please select a smaller range of sets using plot_set')
            else:
                for i,ID in enumerate(Sets):
                    if len(ID) > 15: ID = ID[:15] + '...'
                        
                    buttons_ID.append(dict(type="buttons",
                                        direction="left",
                                        buttons=list([
                                                dict(
                                                    args=[dict(visible=True), [i]],
                                                    args2=[dict(visible=False ), [i]],
                                                    label=f"Show {ID}",
                                                    method="restyle"
                                                    ),
                                            ]),
                                            pad={"r": 0, "t": 0},
                                            showactive=True,
                                            x=0 + 0.25 * int(round((i+1)/(nUnique))),
                                            xanchor="left",
                                            y=(1.2)+ 0.13*(i - nUnique//2 * int(round((i+1)/(nUnique)))) - 0.06*(i - nUnique//2 * int(round((i+1)/(nUnique)))),
                                            #y=1.5 - 0.06*(i - nUnique//2 * int(round((i+1)/(nUnique)))),
                                            yanchor="top"),
                                    )
        
        
        
        # buttons_meta define the metadata variable used to color the traces
        # TODO : this is currently not working for some reson - coming back to it asap
        buttons_meta = [dict(args=[{"marker" : {"color" : f"metadata_colors['{plot_meta}_color_plot']",
                                    #"colorscale" : f"metadata_colors['{plot_meta}_color_map']"
                                    }},
                                    ],
                                    label=f"Meta {plot_meta}",
                                    method="update"
                                    )]
        
        for i, meta in enumerate(metadata.columns):
            if meta == plot_meta: continue
            buttons_meta.append(dict(args=[{"marker" : {"color" : f"metadata_colors['{meta}_color_plot']",
                                    #"colorscale" : f"metadata_colors['{meta}_color_map']"
                                    }},
                                    ],
                                    label=f"Meta {meta}",
                                    method="update"
                                    ))
            
        buttons_meta = [dict(type="dropdown", 
                                direction="down",
                                buttons=buttons_meta,
                                pad={"r": 0, "t": 0},
                                showactive=True,
                                x=0.5,
                                xanchor="left",
                                y=1.074,
                                yanchor="top")]
                              

        # add buttons to figure
        # updatemenus is a list of distinct dictionaries defining 
        # widgets controlling different parameters in the figure 
        fig.update_layout(
            updatemenus=[
                # Markers or Markers + Lines
                dict(
                    type = "buttons",
                    direction = "left",
                    buttons=list([
                        dict(
                            args=[{"mode":['markers' for i in range(nUnique)]}],
                            label="Markers",
                            method="restyle"
                        ),
                        dict(
                            args=[{"mode":['lines+markers' for i in range(nUnique)]}],
                            label="Markers & Lines",
                            method="restyle"
                        ),
                    ]),
                    pad={"r": 10, "t": 10},
                    showactive=True,
                    x=0.1,
                    xanchor="left",
                    y= 1.09,
                    yanchor="top"
                ),
            ] + buttons_ID #+ buttons_meta

            )
            
        # add label
        fig.update_layout(
            annotations=[
                dict(text="<b>Points:", showarrow=False,
                                    x=0, y=1.06, yref="paper", xref="paper", align="left"),
            ]
        )   
        
        
        # define container for widgets
        container = widgets.HBox([meta_widget, save_pdf]) 
        
        # "static" plot
        no_widget_fig = fig
        
        # "non-static" plot 
        fig = go.FigureWidget(fig)
        
    
    # ─── 2d Plot ──────────────────────────────────────────────────────────
    else: # dimension can be either 2 or 3 - if dim == 3: ... else:
        
        # Getting the margins for the plot axes 
        max_x, min_x = max(data[:,0]), min(data[:,0])
        max_y, min_y = max(data[:,1]), min(data[:,1])
        
        # add scatter trace of whole collection
        fig.add_trace(go.Scatter(x=data[:,0],
                                   y=data[:,1],
                                   mode= 'markers',
                                   xaxis='x',
                                   yaxis='y',
                                   visible=False,
                                   showlegend=False, 
                                   marker_color=color_plot,
                                   opacity=0.7,
                                   text=metadata[plot_meta],
                                   marker=dict(
                                   colorscale=col_list,)))
        
        # for each tree_set add a scatter trace to the figure
        Sets, nSetID = np.unique(metadata['SET-ID'], return_counts=True)
        # number of unique tree_sets
        nUnique = len(Sets)
        for i, SetID in enumerate(Sets):
            idx = metadata['SET-ID'] == SetID
            fig.add_trace(go.Scatter(x=data[idx,0],
                                y=data[idx,1],
                                mode= 'markers', 
                                showlegend=False,
                                visible=True,
                                marker_color=metadata_colors[f'{plot_meta}_color_plot'][i],
                                text=metadata['SET-ID'].values[idx],
                                opacity = 0.7,
                                marker=dict(
                                    colorscale=metadata_colors[f'{plot_meta}_color_map'][i],
                                    size = 10,    
                                    line=dict(
                                        #color='MediumPurple',
                                        width=2,
                                    ),
                                ),
                                ))
        
        # ─── Marginal Distribution Of Points ──────────────────────────
        fig.add_trace(go.Histogram(
                y = data[:,1],
                xaxis = 'x2',
                showlegend=False,
                visible=True,
                marker = dict(
                    color = 'rgba(200,200,250,1)'
                )
            ))
        
        fig.add_trace(go.Histogram(
                x=data[:,0],
                yaxis = 'y2',
                showlegend=False,
                visible=True,
                marker = dict(
                    color = 'rgba(200,200,250,1)'
                )
            ))
        
        # ─────────────────────────────────────────────────────────────
            
        # defines the layout of the plot  
        fig.update_layout(
            width=800,
            height=1200,
            autosize=False,
            margin=dict(t=0, b=0, l=0, r=0),
            template="seaborn",
            xaxis= dict(showgrid=True,
                        showline=True,
                        zeroline=True,
                        showticklabels = True,
                        domain = [0,0.85],
                        #range = [min_x - abs(0.1 * min_x), max_x + abs(0.1 * max_x)]
                        ),
            yaxis= dict(showgrid=True,
                        showline=True,
                        zeroline=True,
                        showticklabels = True,
                        domain = [0,0.85],
                        #range = [min_y - abs(0.1 * min_y), max_y + abs(0.1 * max_y)]
                        ),
            plot_bgcolor= 'rgb(240,240,240)',
            xaxis2 = dict(
                zeroline = True,
                showticklabels = True,
                domain = [0.85,1],
                showgrid = True
            ),
            yaxis2 = dict(
                zeroline = True,
                showticklabels = True,
                domain = [0.85,1],
                showgrid = True
            ),
            paper_bgcolor='rgba(0,0,0,0)',
            #hovermode="x unified",
        )
        
        
        buttons_ID = list()
        # buttons_ID are switch on-off buttons that allow to show or hyde specific tree_sets | activated by select
        if select:
            if nUnique > 16:
                warnings.warn('Selection not available for number of sets > 16 - please select a smaller range of sets using plot_set')
            else:
                for i,ID in enumerate(Sets):
                    if len(ID) > 15: ID = ID[:15] + '...'
                        
                    buttons_ID.append(dict(type="buttons",
                                        direction="left",
                                        buttons=list([
                                                dict(
                                                    args=[dict(visible=True), [i + 1]],
                                                    args2=[dict(visible=False ), [i + 1]],
                                                    label=f"Show {ID}",
                                                    method="restyle"
                                                    ),
                                            ]),
                                            pad={"r": 0, "t": 0},
                                            showactive=True,
                                            x=0 + 0.28 * int(round((i+1)/(nUnique))),
                                            xanchor="left",
                                            y=(1.50)+ 0.10*(i - nUnique//2 * int(round((i+1)/(nUnique)))) - 0.04*(i - nUnique//2 * int(round((i+1)/(nUnique)))),
                                            yanchor="top"),
                                    )
        
        # add updatemenus
        fig.update_layout(
            updatemenus=[
                dict(
                    type = "buttons",
                    direction = "left",
                    buttons=list([
                        # change graph from scatterplot to histogram2dconotour plot
                        dict(
                            args=[{"type":  ["scatter"] + ["scatter" for i in range(nUnique)] + ["histogram", "histogram"],
                                   "visible": [False] + [True for i in range(nUnique+2)],},
                                  {"xaxis":{"showgrid":True,
                                                "showline":True,
                                                "zeroline":True,
                                                "showticklabels":True,
                                                "domain":[0,0.85],},
                                   "yaxis":{"showgrid":True,
                                                "showline":True,
                                                "zeroline":True,
                                                "showticklabels":True,
                                                "domain":[0,0.85],},
                                   # xaxis 2 and yaxis2 refer to the second set of axes 
                                   # that are needed to plot the marginal distribution
                                   # the range indicates the portion of total graph space
                                   # allocated to that specific axis
                                   "xaxis2":{"domain":[0.85,1],
                                             "showgrid":True,
                                             "showline":True,
                                             "zeroline":True,
                                             "showticklabels":True,},
                                   "yaxis2":{"domain":[0.85,1],
                                             "showgrid":True,
                                             "showline":True,
                                             "zeroline":True,
                                             "showticklabels":True,},
                                   
                                   "plot_bgcolor":'rgb(240,240,240)',
                                   },
                                  {"skip_invalid" : True,},
                                  ],
                            label="Scatter2D",
                            method="update"
                        ),
                        dict(
                            args=[{"type": ["histogram2dcontour"] + ["scatter" for i in range(nUnique)] + ["histogram", "histogram"], 
                                   "visible": [True] + [False for i in range(nUnique+2)],}, #"showscale": True, "contours":{"coloring":"fill"}
                                  {"xaxis":{"showgrid":True,
                                                "showline":True,
                                                "zeroline":True,
                                                "showticklabels":True,
                                                "domain":[0,1],
                                                },
                                   "yaxis":{"showgrid":True,
                                                "showline":True,
                                                "zeroline":True,
                                                "showticklabels":True,
                                                "domain":[0,1],
                                                },
                                   "xaxis2":{"domain":[0.999,1],
                                             "showgrid":False,
                                                "showline":False,
                                                "zeroline":False,
                                                "showticklabels":False,},
                                   "yaxis2":{"domain":[0.999,1],
                                             "showgrid":False,
                                             "showline":False,
                                             "zeroline":False,
                                             "showticklabels":False,},
                                   "plot_bgcolor":'rgb(240,240,240)',
                                   },
                                  {"skip_invalid" : True,}
                                  ],
                            label="Contour",
                            method="update"
                        )
                    ]),
                    pad={"r": 10, "t": 10},
                    showactive=True,
                    x=0.15,
                    xanchor="left",
                    y=1.3,
                    yanchor="top",
                ),
                # Markers or Markers + Lines
                dict(
                    type = "buttons",
                    direction = "left",
                    buttons=list([
                        dict(
                            args=[{"mode":"markers"},list(range(1, nUnique+1))],
                            label="Markers",
                            method="restyle"
                        ),
                        dict(
                            args=[{"mode":"lines+markers"},list(range(1, nUnique+1))], 
                            label="Markers & Lines",
                            method="restyle"
                        ),
                    ]),
                    pad={"r": 10, "t": 10},
                    showactive=True,
                    x=0.15,
                    xanchor="left",
                    y=1.2,
                    yanchor="top"
                ),
            ] + buttons_ID
        )

        # Add annotation
        fig.update_layout(
            annotations=[
                dict(text="<b>Plot:", showarrow=False,
                                    x=0, y=1.275, yref="paper", xref="paper", align="left"),
                dict(text="<b>Points:", showarrow=False,
                                    x=0, y=1.175, yref="paper", xref="paper", align="left"),
            ]
        )   
        
        # shows points on contour plot
        def response_points(b):
            if fig.data[0].type == "histogram2dcontour":
                if not np.array([fig.data[i].visible for i in range(1, nUnique+1)]).all():
                    for i in range(1, nUnique+1): fig.data[i].visible = True
                else:
                    for i in range(1, nUnique+1): fig.data[i].visible = False

        # widget button -> response_points 
        points_density = widgets.Button(
            description='Show points on Contour plot',
            button_style='info',
            layout= widgets.Layout(width="200px"),)    
        
        # observer meta values (dropdown menu)
        meta_widget.observe(response_meta, names="value")
        # oberver points_density button
        points_density.on_click(response_points)
        
        container = widgets.HBox([meta_widget,points_density, save_pdf]) 
        
        no_widget_fig = fig
        
        fig = go.FigureWidget(fig)
    
    
    # ─────────────────────────────────────────────────────────────────────
    
    # update_point hihglights specific traces on click
    def update_point(trace, points, selector):  
        if points.point_inds:
            traces_start_at = 0
            if dimensions == 2:
                if points.trace_index == 0: return
                traces_start_at = 1
                
            idx_trace = points.trace_index - traces_start_at
            other_idxs = list(range(traces_start_at, nUnique + traces_start_at))
            other_idxs.pop(idx_trace)

            if trace.opacity == 1.0: 
                trace.opacity = 0.7
                for idx in other_idxs: fig.data[idx].opacity = 0.7
            elif trace.opacity == 0.05: trace.opacity = 1.0
            else: 
                trace.opacity = 1.0
                for idx in other_idxs: fig.data[idx].opacity = 0.05
    
    # binds update_point function on every trace
    for trace in fig.data:
            trace.on_click(update_point)
        
    # VBox contains both widgets and the figure
    image = widgets.VBox([container, fig])   
    
    # if save is requested, then a pdf file is returned 
    if save: no_widget_fig.write_html(name_plot + '.html')
    
    # if static is requested, then a less interactive plot is returned
    #! NB: this can be the only way to obtain a plot in cases where 
    #! plotly - ipywidgets - jupyter versions are incompatible
    if static: return no_widget_fig
    # check if we are working in an interactive environment
    # hence ok using widgets
    if hasattr(builtins,'__IPYTHON__'): return image
    else: return no_widget_fig



__author__ = "Andrea Rubbi"
__credits__ = {"Nicola de Maio": 'EMBL-EBI',
    "Myrthe Willemsen": "Freie Universitat Berlin"}
""" maple_RF is the adapted version of the implementation of the Day's algorithm
	for the Robison Foulds computation originally designed by Nicola de Maio for MAPLE.
	Further improvements thanks to Myrthe's work made the implementation more flexible.
	MAPLE code is available on GitHub: https://github.com/NicolaDM/MAPLE"""

import sys, os, warnings, time
import shutil, subprocess, pickle
import multiprocessing as mp
from rich.console import Console
import rich


class Tree(object):
	def __init__(self, name='', children=None, dist=0.000033):
		if name != '': self.name = name
		self.dist = dist
		self.replacements = 0
		self.children = list()
		self.up = None
		self.dirty = True

		if children is not None:
			for child in children:
				self.add_child(child)

	def __repr__(self):
		try: return str(self.name)
		except AttributeError:
			try: return self.name
			except AttributeError: return ""

	def add_child(self, node):
		assert isinstance(node, Tree)
		self.children.append(node)


# function to read input newick string
def readNewick(tree_list, defaultBLen=0.000033, normalizeInputBLen=1.0):
	""" From a list of strings defining phylogenetic trees
        in newick format, returns a list of Tree instances
        containing the biartitions of the tree in a 
        modularized form, suitable for comparisons with other trees.
 
	Args:
		tree_list (list): list of trees in newick format.
		defaultBLen (float, optional): default branch length. Defaults to 0.000033.
		normalizeInputBLen (float, optional): value used to normalize branch lenghts. Defaults to 1.0.

	Returns:
		trees: list of Tree instances
	"""
	trees = list()
	
	tree_list = [tree_list] if type(tree_list) != type(list()) else tree_list
	for tree in tree_list:
		nwString = tree.replace("\n","")
		branchLengthsIn = (":" in nwString)
		index = 0
		node = Tree()
		name = distStr = ""
		finished = False
		while index < len(nwString):
			# start:bipartition
			if nwString[index] == "(":
				newNode = Tree()
				newNode.minorSequences = list()
				node.add_child(newNode)
				newNode.up = node
				node = newNode
				index += 1
			# end:tree
			elif nwString[index] == ";":
				trees.append(node)
				finished = True
				break
			# start
			elif nwString[index] == "[":
				# end
				while nwString[index] != "]":
					index += 1
				index += 1
			# delimiter:branch_length
			elif nwString[index] == ":":
				index += 1
				while nwString[index] != "," and nwString[index] != ")" and nwString[index] != ";":
					distStr += nwString[index]
					index += 1
			# delimiter:nodes
			elif nwString[index] == ",":
				if name != "": node.name, name = name, ""
				if distStr != "":
					node.dist = float(distStr) * normalizeInputBLen
					if node.dist < 0.0:
						warnings.warn(f"Warning: negative branch length in the input tree: {str(distStr)} ; converting it to positive.")
						node.dist = abs(node.dist)
					distStr = ""
				else: node.dist = defaultBLen

				newNode = Tree()
				newNode.minorSequences = list()
				node = node.up
				node.add_child(newNode)
				newNode.up = node
				node = newNode
				index += 1
			# end:bipartition
			elif nwString[index] == ")":
				if name != "": node.name, name = name, ""
				if distStr != "":
					node.dist = float(distStr) * normalizeInputBLen
					distStr = ""
				else:
					node.dist = defaultBLen
				node = node.up
				index += 1
			else:
				name += nwString[index]
				index += 1
    
		assert finished,  "Error, final character ';' not found in newick tree."

	return trees



# Robinson-Foulds distance (1981) using a simplification of the algorithm from Day 1985.
# prepareTreeComparison prepares the data to compare trees to a reference t1.
# RobinsonFouldsWithDay1985(), below, is the implementation of Day's algorithm.

def prepareTreeComparison(t1, rooted=False, minimumBLen=0.000006):
	"""
	For every branch that is shared among two trees,
	add the absolute difference of the two branches to the RFL.
	(This is first stored seperately as the KL value first).
	For every branch that is only in either of the two trees, add the length to the RFL distance.
	For the tree that is 'prepared for comparison', usually the true tree, two hash tables are stored
	(python dictionaries), containing all branch lengths:
		- One contains all leaf-branches. These are indexed by the leaf's name,
  		  usually a number (node.name).
		- The other contains all inner branches, indexed by the <Left, Right> values
  		  of the subtree that is attached by the branch of interest.
	
	Furthermore, it calculates the sum of all the inner branch lengths.
	Whenever in the second function of the Day's algorithm a branch is found in both trees,
	the absolute difference between the two branches is computed, 
	and the result is the difference between that value and the difference with
	the branch length of the true tree.
	This calculation is not performed for leaf branches,
	as we can be sure that a leaf branch exists in both trees (if higher than the minimum length).
	
	Similar to the normal RF distance function, branches with length smaller
	than the minimum branch length are not taken into account.
	However, cases may occur that a branch exists in both trees,
	but in one of them it is lower than the minimum branch length,
	making the RFL less accurate.

	:param t1: input tree;
	:param rooted: set to True if t1 is rooted, default ro False;
	:param minimumBLen: minimum value for branch length, default to 6E-6;
 
	:return: tree metrics, among which the RFL.
	"""
	# dictionary of values given to sequence names
	leafNameDict = dict()
	# list of sequence names sorted according to value
	leafNameDictReverse = list()
	# table containing clusters in the tree
	nodeTable = list()
	# dictionaries with branch lengths
	branchLengthDict, leafDistDict = dict(), dict()
	sumBranchLengths = 0 # sum of all branch lengths, except from the leaves
	# if comparing as unrooted trees, calculate tot num of leaves (using a postorder traversal), which will become useful later
	if not rooted:
		nLeaves = 0
		node = t1
		movingFrom = 0
		while node != t1.up:
			if movingFrom == 0: #0 means reaching node from parent, 1 means coming back from a child
				if len(node.children) == 0:
					nLeaves += 1
					nextNode = node.up
					movingFrom = 1
					nodeTable.append([0,0])
				else:
					nextNode = node.children[0]
					movingFrom = 0
					node.exploredChildren = 0
			else:
				nChildren = len(node.children)
				node.exploredChildren += 1
				if node.exploredChildren == nChildren:
					nextNode = node.up
					movingFrom = 1
				else:
					nextNode = node.children[node.exploredChildren]
					movingFrom = 0
			node = nextNode
			
	# implementing a non-recursive postorder traversal to assign values to internal nodes to fill nodeTable
	leafCount = 0
	node = t1
	movingFrom = 0
	lastL = float("inf")
	lastR = float("-inf")
	lastDesc = 0
	numBranches = 0
	while node != t1.up:
		if movingFrom == 0: #0 means reaching node from parent, 1 means coming back from a child
			if len(node.children) == 0:
				node.name = (node.name).replace("?","_").replace("&","_")
				leafNameDict[node.name] = leafCount
				leafNameDictReverse.append(node.name)
				if rooted:
					nodeTable.append([0,0])
				lastL = leafCount
				lastR = leafCount
				lastDesc = 1
				leafCount += 1
				nextNode = node.up
				movingFrom = 1
				leafDistDict[node.name] = node.dist
			else:
				node.exploredChildren = 0
				node.maxSoFar = float("-inf")
				node.minSoFar = float("inf")
				node.nDescendants = 0
				nextNode=node.children[0]
				movingFrom = 0
		else:
			nChildren = len(node.children)
			node.exploredChildren += 1
			if lastL < node.minSoFar:
				node.minSoFar = lastL
			if lastR > node.maxSoFar:
				node.maxSoFar = lastR
			node.nDescendants += lastDesc
			if node.exploredChildren == nChildren:
				nextNode = node.up
				movingFrom = 1
				lastL = node.minSoFar
				lastR = node.maxSoFar
				lastDesc = node.nDescendants
				if node != t1:
					sumBranchLengths += node.dist
				if node == t1:
					nodeTable[lastR][0] = lastL
					nodeTable[lastR][1] = lastR
				else:
					if (not rooted) and (node.up == t1) and (len(t1.children) == 2):
						if node == t1.children[1]:
							currentBL = node.dist + t1.children[0].dist
							addBranch = True
						else:
							addBranch = False
					else:
						currentBL = node.dist
						addBranch = True
					if addBranch and currentBL > minimumBLen:
						numBranches += 1
						if rooted or lastL > 0:
							if node == node.up.children[-1]:
								if nodeTable[lastL][0] == 0 and nodeTable[lastL][1] == 0:
									nodeTable[lastL][0] = lastL
									nodeTable[lastL][1] = lastR
								else:
									nodeTable[lastR][0] = lastL
									nodeTable[lastR][1] = lastR
							else:
								nodeTable[lastR][0] = lastL
								nodeTable[lastR][1] = lastR
							branchLengthDict[(lastL, lastR)] = currentBL
						else: # re-root at leaf 0, so flip the values for the current branch if it contains leaf 0.
								flippedL = lastR + 1
								flippedR = nLeaves - 1
								nodeTable[flippedL][0] = flippedL
								nodeTable[flippedL][1] = flippedR
								branchLengthDict[(flippedL, flippedR)] = currentBL
			else:
				nextNode = node.children[node.exploredChildren]
				movingFrom = 0
		node = nextNode
	return leafNameDict, nodeTable, leafCount, numBranches, leafDistDict, branchLengthDict, sumBranchLengths

# Robinson-Foulds distance (1981) using a simplification of the algorithm from Day 1985.
# this function compares the current tree t2 to a previous one for which prepareTreeComparison() was run.
# Example usage: leafNameDict, nodeTable, leafCount, numBranches = prepareTreeComparison(phyloTrue,rooted=False)
# numDiffs, normalisedRF, leafCount, foundBranches, missedBranches, notFoundBranches = RobinsonFouldsWithDay1985(phyloEstimated,leafNameDict,nodeTable,leafCount,numBranches,rooted=False)
def RobinsonFouldsWithDay1985(t2, t1, rooted=False, minimumBLen=0.000006):
	leafNameDict,nodeTable,leafCount,numBranches,leafDistDict, branchLengthDict, sumBranchLengths = t1
	# implementing a non-recursive postorder traversal to check branch existance in the reference tree
	node = t2
	# branches in reference tree that are also in t2
	foundBranches = 0
	# branches in t2 that are not found in the reference
	missedBranches = 0
	movingFrom = 0
	lastL = float("inf")
	lastR = float("-inf")
	lastDesc = 0
	visitedLeaves = 0
	RFL = sumBranchLengths
	KF = 0
	while node != t2.up:
		if movingFrom == 0: #0 means reaching node from parent, 1 means coming back from a child
			if len(node.children) == 0:
				node.name = (node.name).replace("?","_").replace("&","_")
				if node.name in leafNameDict:
					leafNum = leafNameDict[node.name]
				else:
					print(node.name + " not in reference tree - aborting RF distance")
					return None, None, None, None, None, None
				lastL = leafNum
				lastR = leafNum
				lastDesc = 1
				nextNode = node.up
				movingFrom = 1
				visitedLeaves += 1
				trueDist = leafDistDict[node.name]
				KF += abs(trueDist - node.dist)   #As described, I have not added branch lengths of leaf nodes to sumBranchLengths, so no need for: RFL=- trueDist
			else:
				node.exploredChildren = 0
				node.maxSoFar = float("-inf")
				node.minSoFar = float("inf")
				node.nDescendants = 0
				nextNode = node.children[0]
				movingFrom = 0
		else:
			nChildren = len(node.children)
			node.exploredChildren += 1
			if lastL < node.minSoFar:
				node.minSoFar = lastL
			if lastR > node.maxSoFar:
				node.maxSoFar = lastR
			node.nDescendants += lastDesc
			if node.exploredChildren == nChildren:
				nextNode = node.up
				movingFrom = 1
				lastL = node.minSoFar
				lastR = node.maxSoFar
				lastDesc = node.nDescendants
				if node != t2:
					if (not rooted) and (node.up == t2) and (len(t2.children) == 2):
						if node == t2.children[1]:
							currentBL = node.dist + t2.children[0].dist
							searchBranch = True
						else:
							searchBranch = False
					else:
						currentBL = node.dist
						searchBranch = True
					if searchBranch and currentBL > minimumBLen:
						if (lastR + 1 - lastL) == lastDesc:
							if rooted or lastL > 0:						
								if nodeTable[lastL][0] == lastL and nodeTable[lastL][1] == lastR:
									foundBranches += 1
									trueDist = branchLengthDict[(lastL, lastR)]
									KF += abs(trueDist - currentBL)
									RFL -= trueDist
								elif nodeTable[lastR][0] == lastL and nodeTable[lastR][1] == lastR:
									foundBranches += 1
									trueDist = branchLengthDict[(lastL, lastR)]
									KF += abs(trueDist - currentBL)
									RFL -= trueDist
								else:
									missedBranches += 1
									RFL += currentBL
							else: # re-root at leaf 0, so flip the values for the current branch if it contains leaf 0.
								flippedL = lastR + 1
								flippedR = leafCount - 1
								if nodeTable[flippedL][0] == flippedL and nodeTable[flippedL][1] == flippedR:
									foundBranches += 1
									trueDist = branchLengthDict[(flippedL, flippedR)]
									KF += abs(trueDist - currentBL)
									RFL -= trueDist
								elif nodeTable[flippedR][0] == flippedL and nodeTable[flippedR][1] == flippedR:
									foundBranches += 1
									trueDist = branchLengthDict[(flippedL, flippedR)]
									KF += abs(trueDist - currentBL)
									RFL -= trueDist
								else:
									missedBranches += 1
									RFL += currentBL
						else:
							missedBranches += 1
							RFL += currentBL
			else:
				nextNode = node.children[node.exploredChildren]
				movingFrom = 0
		node = nextNode
	if visitedLeaves < leafCount:
		print(f"There are leaves in the reference that have not been found in this new tree - leafCount {str(leafCount)} visitedLeaves {str(visitedLeaves)}")
		return None, None, None, None, None, None
	# first value is number of differences, second value is max number of differences just in case one wants the normalized values; 
	# the other values are there just in case one wants more detail.
	numDiffs = ((numBranches-foundBranches) + missedBranches)
	RFL += KF
	if rooted:
		normalization = numBranches + leafCount - 2
	else:
		normalization = numBranches + leafCount - 3
	return numDiffs, float(numDiffs)/(normalization), leafCount, foundBranches, missedBranches, (numBranches - foundBranches), RFL

def calculate_distance_matrix(file, n_trees, output_file):
	console = Console()
	global func_pool
	def func_pool(i):
		return subprocess.check_output(['pypy3', f'{current}/RF_pypy.py', str(i)], universal_newlines = True)
	current = os.path.dirname(os.path.realpath(__file__))	

	with open(file, 'r') as f:
			trees = list(f.read().splitlines())
			f.close()
	distance_matrix = list()
	if shutil.which('pypy3') is not None:
		pckl = open('Trees.pckl', 'wb')
		pickle.dump(trees, pckl)
		pckl.close()

		with mp.Pool() as pool:
			arg_pool = list(range(len(trees)))
			results = list(pool.imap(func_pool, arg_pool))
		distance_matrix = list(map(lambda res: res.split('\n')))
		#for i, tree in enumerate(trees):
		#	status.update(f"[bold red]{i+1}/{len(trees)} [bold blue]Computing Robison Foulds distances...")
		#	command = ['pypy3', f'{current}/RF_pypy.py', str(i)]
		#	res = subprocess.check_output(command, universal_newlines = True).split('\n')
		#	distance_matrix.append(eval(res[1].rstrip()))
	else:
		with console.status(f"[bold green] Robison Foulds by MAPLE") as status:
			time.sleep(1.0)
			for i, tree in enumerate(trees):
						status.update(f"[bold red]{i+1}/{len(trees)} [bold blue]Computing Robison Foulds distances...")
						inputTree = trees[i:i+1]
						inputRFtrees = trees[:] 
						tree1 = readNewick(inputTree)[0]
						tree1_prep = prepareTreeComparison(tree1, rooted=False)
						otherTrees = readNewick(inputRFtrees)
						RF_distances = list()
						for tree in otherTrees:
							res = RobinsonFouldsWithDay1985(tree, tree1_prep, rooted=False)
							RF_distances.append(res[0])
						distance_matrix.append(RF_distances)
	distance_matrix.to_csv(output_file)
	return distance_matrix

        




__author__ = 'Andrea Rubbi'

# ─── About ────────────────────────────────────────────────────────────────────
'''This file defines the tree_set and set_collection classes.
    tree_set contains the information relative to a single set of phyogenetic trees
    in newick format. It allows to compute the distance matrix using different methods and metrics.
    The distance matrix can be then embedded using different methods and subsequently plotted in 2D or 3D.
    A distance matrix and metadata can be given as .csv files. Moreover, metadata is modified
    in order to give information regarding the name of the tree set and the index (or step) of each tree.
    Please note that, once an instance of a class is generated, its metadata dataframe should not be substituted
    as this would invalidate it for the plotting functions. Addition of columns and features is possible by
    accessing the dataframe and modifying it as a pandas.DataFrame instance. 
    set_collection behaves similarly to set_collection. Matter of fact, it is a subclass of the latter and therefore
    shares most of its methods. Its purpose is to analyze concurrently multiple instances of tree_sets and plot their 
    relative distance in a common embedding. Examples of possible applications are present at: ###LINK###'''
# ──────────────────────────────────────────────────────────────────────────────

__copyright__ = "..."
__credits__ = ["Andrea Rubbi", "Lukas Weilguny", "Nick Goldman", "Nicola de Maio"]

__license__ = "..."
__version__ = "1.0.1"
__maintainer__ = "Andrea Rubbi"
__institute__ = 'EMBL-EBI'
__email__ = 'andrear@ebi.ac.uk'
__status__ = "Production"

# ──────────────────────────────────────────────────────────────────────────────
import os, sys, time
import shutil, subprocess
import pandas as pd
import numpy as np
import uuid

#? rich is a very nice library that allows to 
#? easily format the output of console
#? https://github.com/Textualize/rich
from rich.console import Console
from rich import print

# getting the name of the directory
current = os.path.dirname(os.path.realpath(__file__))
 
# Getting the parent directory name
parent = os.path.dirname(current)
 
# adding the parent directory to
# the sys.path.
sys.path.append(parent)

# importing other modules
# if used as a library
try:
    from .embeddings import PCA_e, tSNE_e
    from .calculate_distances import hashrf
    from .calculate_distances import maple_RF
    from .interactive_mode import interactive
    from .embeddings.graph import graph
    from .subsample import subsample
# if used as a program
except:
    from embeddings import PCA_e, tSNE_e
    from calculate_distances import hashrf
    from calculate_distances import maple_RF
    from interactive_mode import interactive
    from embeddings.graph import graph
    from subsample import subsample

# ────────────────────────────────────────────────────────── TREE_SET CLASS ─────
class tree_set():
    """Class for the analysis of a set of phylogenetic trees"""
    # Console from rich -> takes control of console output
    console = Console()
    # ─── INIT ──────────────────────────────────────────────────────────────────
    def __init__(self,
                file, 
                output_file=None,
                distance_matrix=None,
                metadata=None):
        """ Initialize tree_set
        
        file: mandatory - file with set of phylogenetic trees in newick format
        output_file: facultative - specifies output_file of distance matrix
        distance_matrix: facultative - specifies file with already-computed distance matrix
        metadata: facultative - specifies file containing additional information for each tree in set.
        It should contain a column for each feature, a row for each tree (blank row if no info) """
        
        self.file = file
        self.output_file = output_file
        self.distance_matrix = distance_matrix
        self.metadata = metadata
        self.embedding_pca2D = None
        self.embedding_tsne2D = None
        self.embedding_pca3D = None
        self.embedding_tsne3D = None
                 
        if self.output_file == None: self.output_file = "./{file}_distance_matrix.csv".format(file = os.path.splitext(os.path.basename(self.file))[0])
        
        with open(file, 'r') as f:
            self.n_trees = len(f.readlines())
            f.close()
        
        if type(self.distance_matrix) != type(None):
            try: pd.read_csv(self.distance_matrix) 
            except: print("There's an error with the Distance Matrix file - please check the correct location and name of the .csv file"), exit() 
              
        if type(self.metadata) != type(None):
            try: self.metadata = pd.read_csv(self.metadata) 
            except: print("There's an error with the Metadata file - please check the correct location and name of the .csv file"), exit() 
        
        else: self.metadata = pd.DataFrame()
        self.metadata['SET-ID'] = [os.path.splitext(os.path.basename(self.file))[0] for i in range(self.n_trees)]
        self.metadata['STEP'] = [i for i in range(self.n_trees)]
        self.sets = np.unique(self.metadata['SET-ID'])
            
    # ─── STR ───────────────────────────────────────────────────────────────────
    def __str__(self):
        computed = 'not computed'
        if type(self.distance_matrix) != type(None): computed = 'computed'
            
        return(f"─────────────────────────────\n Tree set containing {self.n_trees} trees;\n File: {self.file};\n Distance matrix: {computed}.\n───────────────────────────── \n")
    
    # ─── CALCULATE DISTANCES ───────────────────────────────────────────────────
    def calculate_distances(self, method):
        methods = {'hashrf' : hashrf.hashrf,
                   'hashrf_weighted' : hashrf.hashrf_weighted,
                   'days_RF' : maple_RF.calculate_distance_matrix,
                   'None' : None, }

        with self.console.status("[bold green]Calculating distances...") as status:
            self.distance_matrix = methods[method](self.file, self.n_trees, self.output_file)
        print(f'[bold red]{method} | Done!')
    
    # ─── EMBED ─────────────────────────────────────────────────────────────────
    def embed(self, method, dimensions):
        methods = {'pca' : PCA_e.pca,
                   'tsne' : tSNE_e.tsne,
                   'None' : None, }
        
        if dimensions == 2:
        
            if type(self.distance_matrix) == type(None): self.calculate_distances('hashrf')
                #raise Exception('Distance matrix has to be computed or inserted by the user prior the embedding')
            if method == 'pca':
                self.embedding_pca2D = methods[method](self.distance_matrix, dimensions, self.metadata)
            if method == 'tsne':
                self.embedding_tsne2D = methods[method](self.distance_matrix, dimensions, self.metadata)
        
        if dimensions == 3:
        
            if type(self.distance_matrix) == type(None): self.calculate_distances('hashrf') 
                #raise Exception('Distance matrix has to be computed or inserted by the user prior the embedding')
            if method == 'pca':
                self.embedding_pca3D = methods[method](self.distance_matrix, dimensions, self.metadata)
            if method == 'tsne':
                self.embedding_tsne3D = methods[method](self.distance_matrix, dimensions, self.metadata)        
    
    # ─── PLOT EMBEDDING ─────────────────────────────────────────────────────────
    
    def plot_2D(self, method, save=False, name_plot=None, static=False, plot_meta = 'SET-ID', plot_set = None, select = False, same_scale = False):
        if type(plot_set) == type(None): plot_set = self.sets
        if method == 'pca':
            if name_plot == None: name_plot='PCA_2D'
            if type(self.embedding_pca2D) == type(None): self.embed('pca', 2)
            fig = graph.plot_embedding(self.embedding_pca2D, self.metadata, 2, save, name_plot, static, plot_meta, plot_set, select, same_scale)
        
        elif method == 'tsne':
            if name_plot == None: name_plot='TSNE_2D'
            if type(self.embedding_tsne2D) == type(None): self.embed('tsne', 2)
            fig = graph.plot_embedding(self.embedding_tsne2D, self.metadata, 2, save, name_plot, static, plot_meta, plot_set, select, same_scale)
        
        else: raise ValueError("'method' can only be either 'pca' or 'tsne' ")
        
        return fig
    
    def plot_3D(self, method, save=False, name_plot=None, static=False, plot_meta = 'SET-ID', plot_set = None, select = False, same_scale = False):
        if type(plot_set) == type(None): plot_set = self.sets
        if method == 'pca':
            if name_plot == None: name_plot='PCA_3D'
            if type(self.embedding_pca3D) == type(None): self.embed('pca', 3)
            fig = graph.plot_embedding(self.embedding_pca3D, self.metadata, 3, save, name_plot, static, plot_meta, plot_set, select, same_scale)
        
        elif method == 'tsne':
            if name_plot == None: name_plot='TSNE_3D'
            if type(self.embedding_tsne3D) == type(None): self.embed('tsne', 3)
            fig = graph.plot_embedding(self.embedding_tsne3D, self.metadata, 3, save, name_plot, static, plot_meta, plot_set, select, same_scale)
        
        else: raise ValueError("'method' can only be either 'pca' or 'tsne' ")
        
        return fig
    
    # ─── GET SUBSET ───────────────────────────────────────────────────────
    
    def get_subset(self, n_required, T = 1):
        console = Console()
        with console.status("[bold blue]Extracting subsample...") as status:
            if shutil.which('pypy3') is not None: 
                command = ['pypy3', f'{current}/subsample/subsample.py', self.file, str(self.n_trees), str(n_required)]
                res = subprocess.check_output(command, universal_newlines = True).split('\n')
                subsample_trees, idxs = eval(res[3]), eval(res[4])
            else:
                console.log('[bold red]Could not find pypy3 on your sytem PATH - using python3...')                
                subsample_trees, idxs = subsample.subsample(self.file, self.n_trees, n_required, subp = False)
            
            file_sub = f'SUBSAMPLE'
            with open(file_sub, 'w') as f:
                for i in subsample_trees: f.write(i)
                f.close()
                
            status.update("[bold green]Calculating distances...")
            dM = hashrf.hashrf(file_sub, n_required, file_sub + '_distances.csv')
            components = PCA_e.pca(dM, 3)
            status.update(f'[bold blue] Done!')
            time.sleep(0.2)
        
        sorted_elements = sorted(enumerate(idxs), key=lambda x:x[1])
        idxs_sorted, order = list(map(lambda x:x[1], sorted_elements)), list(map(lambda x:x[0], sorted_elements))   
        comp_sorted = np.array([components[line,:] for line in order])
        #dM_sorted = np.array([dM[line,:] for line in order])
        SetID_sub = self.metadata['SET-ID'][idxs_sorted]
        meta_sub = pd.DataFrame({'SET-ID':SetID_sub, 'STEP': idxs_sorted})       
        
        return graph.plot_embedding(comp_sorted, meta_sub, 2), graph.plot_embedding(comp_sorted, meta_sub, 3)

            
# ──────────────────────────────────────────────────────────────────────────────
# ─────────────────────────────────────────────────── SET_COLLECTION CLASS ─────
class set_collection(tree_set):
    # NB: set_collection is a sub_class of tree_set
    # therefore, most methods are shared between these two classes
    def __init__(self, collection = list(), file = 'Set_collection', 
                 output_file = "./Set_collection_distance_matrix",
                 distance_matrix = None):
        """ Initialize set_collection
        
        collection: facultative - tree_set or list of tree_sets
        NB: if no collection is given an empty set_collection is generated 
        file: facultative - file with set of phylogenetic trees in newick format
        output_file: facultative - specifies output_file of distance matrix"""
        
        self.id = uuid.uuid4()
        self.file = file + str(self.id)
        self.distance_matrix = pd.read_csv(distance_matrix, header=None).values if distance_matrix else distance_matrix
        self.embedding_pca2D = None
        self.embedding_tsne2D = None
        self.embedding_pca3D = None
        self.embedding_tsne3D = None
        
        if self.file != 'Set_collection' + str(self.id) and output_file == "./Set_collection_distance_matrix.csv": 
            self.output_file = "./{file}_distance_matrix.csv".format(file = os.path.splitext(os.path.basename(self.file))[0])
        else: self.output_file = output_file + str(self.id) + ".csv"
        
        if isinstance(collection, tree_set): 
            self.collection = [collection]
            with open(self.file, 'w') as trees:
                with open(collection.file, 'r') as file:
                        trees.write(file.read())
                        file.close()
                trees.close()
            
        elif len(collection) > 0:
            for element in collection: assert isinstance(element, tree_set), 'Every element in a set_collection must be a tree_set'
            self.collection = collection
            with open(self.file, 'w') as trees:
                for set in collection:
                    with open(set.file, 'r') as file:
                        trees.write(file.read())
                        file.close()
                trees.close()
        
        else: self.collection = collection
        
        with open(self.file, 'r') as f:
            self.n_trees = len(f.readlines())
            f.close()
            
        self.data = dict()
        
        self.metadata = pd.DataFrame()
        
        for set in self.collection:
            key = os.path.splitext(os.path.basename(set.file))[0]
            
            metadata = set.metadata
            if type(metadata) == type(None): metadata = pd.DataFrame()
            metadata['SET-ID'] = np.array([key] * set.n_trees)
            
            self.metadata = pd.concat([self.metadata, metadata])
            
            self.data[key] = {'metadata':metadata, 
                                 'n_trees':set.n_trees}
        
        self.metadata.reset_index(drop=True, inplace=True)
        
        self.sets = np.unique(self.metadata['SET-ID'])
            
        
    # the result of addition between two collections 
    # is the concatenation of the two collections
    def __add__(self, other):
        try: 
            assert isinstance(other, tree_set)
            return set_collection(self.collection + [other])
            
        except: 
            try: 
                assert isinstance(other, set_collection)
                return set_collection(self.collection + other.collection)
            except: 
                for element in other: assert isinstance(element, tree_set), 'You can concatenate a set_collection only with another set_collection, a tree_set, or a list of tree_set'
                return set_collection(self.collection + other)
                
    def __str__(self):
        computed = 'not computed'
        if type(self.distance_matrix) != type(None): computed = 'computed'
            
        summary = f"─────────────────────────────\n Tree set collection containing {self.n_trees} trees;\n File: {self.file};\n Distance matrix: {computed}.\n───────────────────────────── \n"
        for key,value in self.data.items():
            summary += f"{key}; Containing {value['n_trees']} trees. \n" 
        
        return summary
        
    # concatenate is a more formal method to concatenate collections
    # using this allows for more clarity in the codebase
    #! less error prone !#
    def concatenate(self, other):
        try: 
            assert isinstance(other, tree_set)
            return set_collection(self.collection + [other])
            
        except: 
            try: 
                assert isinstance(other, set_collection)
                return set_collection(self.collection + other.collection)
            except: 
                for element in other: assert isinstance(element, tree_set), 'You can concatenate a set_collection only with another set_collection, a tree_set, or a list of tree_set'
                return set_collection(self.collection + other)
            
            
            
        
        

         